/* 
 * file-documentation
 * =========================================================================
 * alloc.h -- Memory allocation macro's
 * 
 * Author          : Hans Fleurkens
 * Created On      : Tue Jan 16 14:11:46 1990
 * Last Modified By: Hans Fleurkens
 * Last Modified On: Thu Sep  5 10:26:32 1991
 * Update Count    : 34
 * Status          : OK
 * 	
 * History
 * Description:
 *        This module contains macros to allocate and free dynamic memory.
 *        These routines can be used, if the module 'error.c' is linked
 *        with the application.
 * =========================================================================
 * Copyright (c) 1990, 1991
 * Hans Fleurkens
 * Eindhoven University of Technology
 * The Netherlands
 * =========================================================================
 */


#ifndef ALLOC_H
#define ALLOC_H

#include <stdlib.h>

extern void *calloc_ptr; /* defined in Error.c */
extern char *str_ptr;

extern void errorMsg ();

#define NEW_STRUCT(type) \
  (type *) ((calloc_ptr = calloc (1, sizeof (type))) \
	    ? calloc_ptr \
	    : (void *)(errorMsg("Memory allocation failed"), \
		       exit(1),NULL))

#define NEW_ARRAY(n,type) \
  (type *) ((calloc_ptr = calloc (n, sizeof (type))) \
	    ? calloc_ptr \
	    : (void *)(errorMsg("Memory allocation failed"), \
		       exit(1),NULL))

#define NEW_STRING(cp) \
  ((str_ptr = calloc (strlen (cp)+1, sizeof (char))) \
   ? strcpy (str_ptr, cp) \
   : (char *)(errorMsg("Memory allocation failed"), \
	      exit(1),NULL))

#define REALLOC(ptr,n,type) \
  (type *)((calloc_ptr=realloc(ptr,n*sizeof(type))) \
	   ? calloc_ptr \
	   : (void *)(errorMsg("Memory allocation failed"), \
		      exit(1),NULL))

/* Allocates n consecutive bytes, returns pointer of type. */
#define MALLOC_BYTES(n, type) \
  (type) ((calloc_ptr = malloc (n)) \
	    ? calloc_ptr \
	    : (void *)(errorMsg( "Cannot perform alloc...."), \
		       exit(1),NULL))

/* Allocates n consecutive bytes (zeroed), returns pointer of type. */
#define CALLOC_BYTES(n, type) \
  (type) ((calloc_ptr = calloc (n, sizeof (char))) \
	    ? calloc_ptr \
	    : (void *)(errorMsg( "Cannot perform alloc...."), \
		       exit(1),NULL))

/* Reallocates p, to point to a stretch of n bytes. */
#define REALLOC_BYTES(p, n, type) \
  (type) ((calloc_ptr = realloc (p, (n) * sizeof (char))) \
	    ? calloc_ptr \
	    : (void *)(errorMsg( "Cannot perform alloc...."), \
		       exit(1),NULL))

/* Allocates n elements of type. */
#define MALLOC_ARRAY(n, type) \
  (type *) ((calloc_ptr = malloc ((n) * sizeof (type))) \
	    ? calloc_ptr \
	    : (void *)(errorMsg( "Cannot perform alloc...."), \
		       exit(1),NULL))

/* Allocates n elements of type, all elements zeroed. */
#define CALLOC_ARRAY(n, type) \
  (type *) ((calloc_ptr = calloc (n, sizeof (type))) \
	    ? calloc_ptr \
	    : (void *)(errorMsg( "Cannot perform alloc...."), \
		       exit(1),NULL))

/* Reallocates array of type to n elements. */
#define REALLOC_ARRAY(arr, n, type) \
  (type *) ((calloc_ptr = realloc (arr, (n) * sizeof (type))) \
	    ? calloc_ptr \
	    : (void *) (errorMsg( "Cannot perform alloc...."), \
			exit(1),NULL))

/* Allocates string of len chars (takes care of extra null-terminator byte). */
#define CALLOC_STRING(len) \
  ((str_ptr = calloc ((len)+1, sizeof (char))) \
   ? str_ptr \
   : (char *) (errorMsg( "Cannot perform alloc...."), \
	       exit(1),NULL))

#define COPY_STRING_1(cp, len) \
  ((str_ptr = malloc ((len)+1)) \
   ? (strncpy (str_ptr, cp, len), \
      str_ptr[len]='\0', str_ptr) \
   : (char *) (errorMsg( "Cannot perform alloc...."), \
	       exit(1),NULL)) 

/* Allocates struct of type. */
#define MALLOC_STRUCT(type) \
  (type *) ((calloc_ptr = malloc (sizeof (type))) \
	    ? calloc_ptr \
	    : (void *) (errorMsg( "Cannot perform alloc...."), \
			exit(1),NULL))

/* Allocates struct of type, all fields zeroed. */
#define CALLOC_STRUCT(type) \
  (type *) ((calloc_ptr = calloc (1, sizeof (type))) \
	    ? calloc_ptr \
	    : (void *) (errorMsg( "Cannot perform alloc...."), \
			exit(1),NULL))

#define FREE(ptr) 	{ if (ptr) {free((void *)ptr); ptr = NULL;} }

#endif /* ALLOC_H */
