/*
 DOCUMENTATION INFORMATION				    module: HASH TABLE
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 system	   : Apollo DN3000, HP 9000/S500
 file	   : hash.c
 unit-title: 
 ref.	   : 
 author(s) : Copyright (c) 1988, 1989, 1990 G.L.J.M. Janssen
             1990 Hans Fleurkens
 date	   :  8-JAN-1990
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

#include <stdio.h>
#include <string.h>

#include "alloc.h"
#include "hash.h"

/* define this to cause debugging code to be compiled: */
/* #define CC_DEBUG*/

#define X_EMPTY_BUCKET(entries,h)	(!entries[h])
#define X_OCCUPIED_BUCKET(entries,h)	(entries[h])
#define X_KEYLEN(entries,h)		entries[h]->keylen
#define X_KEYSTR(entries,h)		entries[h]->keystr
#define X_KEYINFO(entries,h)		entries[h]->info

HASHTAB_ENTRY_PTR all_hashtab_entries = NULL;
HASHTAB_ENTRY     null_hashtab_entry = {0};
HASHTAB_ENTRY_PTR temp_hashtab_entry;

/* A table of about 100,000 entries should be enough for most applications
   we foresee now, so no checks are done on overflow. Feel free to extend
   when necessary by adding more primes.
*/

static int primes[] =
{
/*  0 */      13,
/*  1 */      23,
/*  2 */      53,	      
/*  3 */     101,
/*  4 */     199,
/*  5 */     401,
/*  6 */     797,
/*  7 */    1601,
/*  8 */    3203,
/*  9 */    6397,
/* 10 */   12799,
/* 11 */   25601,
/* 12 */   51199,
/* 13 */  100003
};

static int copy_the_entry = 1;
static int lookup_1 ();
/* Run-time statistics and messages. */
int RT_DEBUG = 0;

/* HASHTAB make_hashtab (index);

   Creates a hash table of about 100 * (2 ^ index) entries.
   Min. value for index is 0;
   Max. value for index is currently 10.
*/
HASHTAB make_hashtab (index)
     int index;
{
  HASHTAB tab;
  int size = primes[index];

/* Current maximum size (i.e. number of entries) of hash table: */
  tab.size          = size;

/* This reflects the number of items currently held by the hash table: */
  tab.nr_items      = 0;

/* This counts the number of times an insert of an item is done, i.e.
   number of times `lookup (,,, INSERT)' is called. */
  tab.nr_inserts    = 0;

/* We count as collisions the fact that when an item has to be inserted
   the entry at its hash index is occupied by a different one. Subsequent
   tries to resolve this collision are not counted as new collisions.
   Merely looking up an item will never increase the collision count.
*/
  tab.nr_collisions = 0;

/* Number of rehashes undertaken for this table: */
  tab.nr_rehashes   = 0;

/* Current index in primes table: */
  tab.primes_index  = index;


/* Function called when non-NULL to process change of entry
   when rehashing in process.
*/
  tab.rehash_function = 0;

/* Allocate an array of empty entries: */
  tab.entries       = CALLOC_ARRAY (size, HASHTAB_ENTRY_PTR);

  return tab;
}

void reinit_hashtab (tab)
  HASHTAB *tab;
{
  register int hashtab_size = tab->size;

  /* Free all entries allocated by hash table routines: */
  while (hashtab_size--)
    free_hashtab_entry (tab, hashtab_size);

  tab->nr_inserts    = 0;
  tab->nr_collisions = 0;
  tab->nr_rehashes   = 0;
  /* keep assigned rehash_function! */
}

#ifdef CC_DEBUG
static void print_hashtab (tab)
     HASHTAB tab;
{
  register int i;
  int hashtab_size = tab.size;
  register HASHTAB_ENTRY_PTR *entries = tab.entries;

  for (i = 0; i < hashtab_size; i++) {
    if (X_EMPTY_BUCKET (entries, i)) {
    }
    else {
      /* i: len: contents: */
      printf ("%d %d %s\n", i, X_KEYLEN (entries, i), X_KEYSTR (entries, i));
    }
  }
  printf ("Number of items: %d\n", tab.nr_items);
  printf ("Inserts   total: %d\n", tab.nr_inserts);
  printf ("Collision total: %d\n", tab.nr_collisions);
  printf ("Nr. of rehashes: %d\n", tab.nr_rehashes);
}
#endif

/* This is the hash function. */
static int hash (tab, s, len)
     HASHTAB tab;
     char *s;
     int len;
{
  register char *p = s;
  register char *end = p + len;
  register char c;
  register int hash = 0;

  while (p != end) {
    c = *p++;
    if (c >= 0140) c -= 40;	/* 0140 = ' */
    hash = ((hash<<3) + (hash>>28) + c);
  }
  return (hash & 0x7FFFFFFF) % tab.size;
}

static HASHTAB hashtab_full_handler (tab)
     HASHTAB tab;
{
#ifdef ALLOW_REHASH
  int size = tab.size;
  register HASHTAB_ENTRY_PTR *entries = tab.entries;
  HASHTAB newtab;
  int newsize;
  register int i, newi;
  int insert_var;

  if (RT_DEBUG)
    fprintf (stderr, "Rehashing (old size: %d)...\n", size);

  tab.primes_index++;
  newtab  = tab;
  newsize = primes[newtab.primes_index];
  newtab.entries  = CALLOC_ARRAY (newsize, HASHTAB_ENTRY_PTR);
  newtab.size     = newsize;
  newtab.nr_items = 0;
  newtab.nr_rehashes++;

  copy_the_entry = 0;
  for (i = 0; i < size; i++)
    if (tab.entries[i]) {
      /* Entry i is occupied; insert it in new table: */
      insert_var = INSERT;
      /* lookup_1 returns index of this entry in newtab: */
      newi = lookup_1 (&newtab, entries[i], &insert_var);
      /* Call user supplied function to account for change: */
      if (tab.rehash_function)
	(*(tab.rehash_function)) (newtab.entries[newi], newi);
    } /*if-for*/

  /* No longer need entries-array of old table: */
  free (entries);
  copy_the_entry = 1;

  if (RT_DEBUG)
    fprintf (stderr, "Rehashing (new size: %d)...done\n", newsize);

  return newtab;
#else
  errorMsg ("Hashtable full; program aborted.");

#ifdef CC_DEBUG
  print_hashtab ();
#endif

  exit (1);
#endif
}

static int lookup_1 (tab, entry, do_insert)
     HASHTAB *tab;		/* the table */
     HASHTAB_ENTRY_PTR entry;	/* the entry to look for */
     int *do_insert;		/* insert it? */
{
  HASHTAB_ENTRY_PTR *hashtab;
  int hashtab_size;
  register int d;
  register int h;
  register char *s = entry->keystr;
  register int len = entry->keylen;
  int first_time  = 1;
  int insert_flag = *do_insert;

  if (insert_flag) tab->nr_inserts++;

 restart:
  hashtab      = tab->entries;
  hashtab_size = tab->size;

  h = hash (*tab, s, len);
  d = 1;

  while (1) {
    if (X_EMPTY_BUCKET (hashtab, h)) {
      if (insert_flag) {

	tab->nr_items++;

	if (copy_the_entry) {
	  char *p;

	  hashtab[h] = CALLOC_HASHTAB_ENTRY ();
	  /* CALLOC_STRING includes terminating null char! */
	  p = X_KEYSTR (hashtab, h) = CALLOC_STRING (len);
	  bcopy (s, p, len);
	  /* p[len] = '\0'; CALLOC_STRING zeroes */
	  X_KEYLEN  (hashtab, h) = len;
	  X_KEYINFO (hashtab, h) = entry->info;
	}
	else
	  hashtab[h] = entry;

	*do_insert = INDEED_INSERTED;
	return h;
      }
      return (NOT_PRESENT);
    }

    /* here: X_OCCUPIED_BUCKET (h) */
    if (   (len == X_KEYLEN (hashtab, h))
	&& (!bcmp (s, X_KEYSTR (hashtab, h), len))) {
      *do_insert  = ALREADY_PRESENT;
      entry->info = X_KEYINFO (hashtab, h);
      return h;
    }

    /* here: COLLISION */
    if (insert_flag && first_time) {
      tab->nr_collisions++;
      first_time--;
    }

    if (d == hashtab_size) {
      if (!insert_flag) return (NOT_PRESENT);

      *tab = hashtab_full_handler (*tab);
      goto restart;
    }

    /* quadratic probing to find free place */
    h += d;
    d += 2;
    if (h >= hashtab_size) h -= hashtab_size;	     
  } /*while (1)*/
}

/* int lookup (tab, s, len, info, do_insert)

   Looks up string `s' of length `len' in hash table `tab'
   (must be address of HASHTAB variable).
   String need not be null terminated and may contain null characters.
   Returns index if found else `NOT_PRESENT'.

   `do_insert' can be one of the constants LOOKUP or INSERT, or the
   address of an int variable containing one of those constants.
   LOOKUP means just see if `s' is already present.
   INSERT means insert the string when not already present. In case of
   a variable, ALREADY_PRESENT or INSERTED is reported back.
   String is copied in that case and null terminator character is added.

   When a new item is inserted it is associated with the `info', which
   must be the address of character pointer variable upon entry, or NULL
   if info is of no conceren.
   When an item is retrieved, i.e. it was already present, its info is returned
   through `info'.

   Uses quadratic probing to resolve collisions.
   Calls `hashtab_full_handler' upon full hash table.
*/
int lookup (tab, s, len, info, do_insert)
     HASHTAB *tab;
     char *s;
     int len;
     char **info;
     int *do_insert;
{
  HASHTAB_ENTRY entry;
  int insert_var;
  int report_back = 0;
  int index;

  /* fill entry for lookup_1: */
  entry.keystr = s;
  entry.keylen = len;
  entry.info   = info ? *info : NULL;

  if (	 do_insert != (int *) LOOKUP
      && do_insert != (int *) INSERT) {
    insert_var = *do_insert;
    report_back++;
  }
  else
    insert_var = (int) do_insert;

  index = lookup_1 (tab, &entry, &insert_var);
  if (info) *info = entry.info;
  if (report_back) *do_insert = insert_var;
  return index;
}

/* void free_hashtab_entry (tab, index)

   Frees the entry at `index' in hash table `tab'.
   (must be address of HASHTAB variable).
   Afterwards EMPTY_BUCKET (tab.entries, index) is true.
*/
void free_hashtab_entry (tab, index)
     HASHTAB *tab;
     int index;
{
  register HASHTAB_ENTRY_PTR *entries = tab->entries;

  if (X_OCCUPIED_BUCKET (entries, index)) {
    tab->nr_items--;
    free (X_KEYSTR (entries, index));
    FREE_HASHTAB_ENTRY (entries[index]);
    tab->entries[index] = (HASHTAB_ENTRY_PTR) 0;
  }
}

/* Frees all space allocated by hash tab routines, but does not reinitialise
   it.
*/
void free_hashtab (tab)
     HASHTAB tab;
{
  register int hashtab_size = tab.size;

  while (hashtab_size--)
    free_hashtab_entry (&tab, hashtab_size);

  free (tab.entries);
}

/* Change this string if you prefer other prefixes: */
static char *prefix = "X_";
static int gen_counter = 0;

/* int gen_unique_entry (tab)

   Generates a entry in the hash table `tab' (must be address of HASHTAB
   variable) with a name that is unique w.r.t. the names already contained
   in the hash table.

   Names generated look like "X_<number>".
   Inserts the new name, clears the info field and returns the hash index.
*/
int gen_unique_entry (tab)
     HASHTAB *tab;
{
  char buf[16], *startp;
  register int i, len = strlen (prefix);
  int inserted;

  strcpy (buf, prefix);
  startp = &buf[len];

  do {
    sprintf (startp, "%d", gen_counter++);
    len = strlen (buf);
    inserted = INSERT;
    i = lookup (tab, buf, len, NULL, &inserted);
  } while (inserted == ALREADY_PRESENT);

  /* here: inserted == INDEED_INSERTED, so i is the right index. */
  return i;
}

#ifdef CC_DEBUG
main (argc, argv)
     int argc;
     char* argv[];
{
  char s[256];
  int lc = 0;
  int i;
  HASHTAB hashtab;
  HASHTAB hashtab2;

  RT_DEBUG = 1;

  hashtab  = make_hashtab ();
  hashtab2 = make_hashtab ();

  i = gen_unique_entry (&hashtab);
  free_hashtab_entry (&hashtab, i);
  i = gen_unique_entry (&hashtab);
  free_hashtab_entry (&hashtab, i);
  i = gen_unique_entry (&hashtab);
  free_hashtab_entry (&hashtab, i);

  while (gets (s)) {
    lc++;
    if (lc & 1)
      lookup (&hashtab , s, strlen (s), NULL, INSERT);
    else
      lookup (&hashtab2, s, strlen (s), NULL, INSERT);
  }
  print_hashtab (hashtab);
  print_hashtab (hashtab2);

  printf ("Nr. input lines: %d\n", lc);

  free_hashtab (hashtab2);
  free_hashtab (hashtab);
}
#endif
