/*
 DOCUMENTATION INFORMATION				    module: HASH TABLE
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 system	   : Apollo DN3000, HP 9000/S500
 file	   : hash.h
 unit-title: 
 ref.	   : 
 author(s) : Copyright (c) 1988, 1989, 1990 G.L.J.M. Janssen
 date	   :  8-JAN-1990
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

/* Include file for hashtable module. */
/* Declares all names needed to use it. */

/* Define this to allow rehashing upon table full, else an error
   message is printed on stderr and the program is aborted.
   Default table size is 101 entries, and is about doubled on each rehash.
*/
#define ALLOW_REHASH

typedef struct HASHTAB_ENTRY {
  int	keylen;			/* length of keystr field */
  char *keystr;			/* the string key */
  char *info;			/* general field for extra info */
} HASHTAB_ENTRY, *HASHTAB_ENTRY_PTR;

/* Note: a hash table is a C struct. Its entries field is an (dynamic) array
   of pointers to entry structs.
*/
typedef struct HASHTAB {
  int size;			/* size of entries array */
  int nr_items;			/* nr. items contained */
  int nr_inserts;		/* nr. attempted inserts */
  int nr_collisions;		/* nr. collisions */
  int nr_rehashes;		/* nr. rehashes */
  int primes_index;
#ifdef ALLOW_REHASH
  void (*rehash_function) ();	/* function called when non-NULL */
				/* to process change of entry */
				/* when rehashing in process. */
				/* Called as: */
  /* rehash_function (tab.entries[new_index], new_index); */
				/* i.e. you get a pointer to the entry */
				/* of type HASHTAB_ENTRY_PTR and the */
				/* index of that entry in the rehashed table.*/
#endif
  HASHTAB_ENTRY_PTR *entries;
} HASHTAB;

#define HASHTAB_SIZE(tab)		(tab.size)

#ifdef COMMENT
#define CALLOC_HASHTAB_ENTRY()		CALLOC_STRUCT (HASHTAB_ENTRY)
#define FREE_HASHTAB_ENTRY(h)		FREE (h)
#endif

extern HASHTAB_ENTRY_PTR all_hashtab_entries;
extern HASHTAB_ENTRY     null_hashtab_entry;
extern HASHTAB_ENTRY_PTR temp_hashtab_entry;

#define CALLOC_HASHTAB_ENTRY() \
	( \
	 all_hashtab_entries ? (temp_hashtab_entry = all_hashtab_entries, \
           all_hashtab_entries = \
	      (HASHTAB_ENTRY_PTR) (all_hashtab_entries->info), \
	   *temp_hashtab_entry = null_hashtab_entry, \
	   temp_hashtab_entry) \
	 : CALLOC_STRUCT (HASHTAB_ENTRY) \
	)

#define FREE_HASHTAB_ENTRY(c) \
	{ \
	  HASHTAB_ENTRY_PTR _xyz_c = c; \
	  \
          _xyz_c->info = (char *) all_hashtab_entries; \
	  all_hashtab_entries = _xyz_c; \
	}

/* Useful symbols for values of do_insert in lookup (). */
#define LOOKUP 0
#define INSERT 1

/* Symbols for values returned by do_insert in lookup (). */
#define ALREADY_PRESENT 0
#define INDEED_INSERTED	1

/* Returned by lookup when string not present in table upon LOOKUP: */
#define NOT_PRESENT -1

#define EMPTY_BUCKET(tab,h)		(!tab.entries[h])
#define OCCUPIED_BUCKET(tab,h)		(tab.entries[h])
#define KEYLEN(tab,h)			tab.entries[h]->keylen
#define KEYSTR(tab,h)			tab.entries[h]->keystr
#define KEYINFO(tab,h)			tab.entries[h]->info

/* When set, causes some informative messages to be send to stderr. */
extern int RT_DEBUG; /* = 0 */

/* HASHTAB make_hashtab (index);

   Creates a hash table of about 100 * (2 ^ index) entries.
   Min. value for index is 0;
   Max. value for index is currently 10.
*/
extern HASHTAB make_hashtab ();

/* int lookup (tab, s, len, info, do_insert)

   Looks up string `s' of length `len' in hash table `tab'
   (must be address of HASHTAB variable).
   String need not be null terminated and may contain null characters.
   Returns index if found else `NOT_PRESENT'.

   `do_insert' can be one of the constants LOOKUP or INSERT, or the
   address of an int variable containing one of those constants.
   LOOKUP means just see if `s' is already present.
   INSERT means insert the string when not already present. In case of
   a variable, ALREADY_PRESENT or INSERTED is reported back.
   String is copied in that case and null terminator character is added.

   When a new item is inserted it is associated with the `info', which
   must be the address of character pointer variable upon entry, or NULL
   if info is of no conceren.
   When an item is retrieved, i.e. it was already present, its info is returned
   through `info'.

   Uses quadratic probing to resolve collisions.
   Calls `hashtab_full_handler' upon full hash table.
*/
extern int lookup ();

/* void free_hashtab_entry (tab, index)

   Frees the entry at `index' in hash table `tab'.
   (must be address of HASHTAB variable).
   Afterwards EMPTY_BUCKET (tab.entries, index) is true.
*/
extern void free_hashtab_entry ();

extern void free_hashtab ();

extern void reinit_hashtab ();

/* int gen_unique_entry (tab)

   Generates a entry in the hash table `tab' (must be address of HASHTAB
   variable) with a name that is unique w.r.t. the names already contained
   in the hash table.

   Names generated look like "X_<number>".
   Inserts the new name, clears the info field and returns the hash index.
*/
extern int gen_unique_entry ();
