#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

#include <Xm/Xm.h>

#include "defines.h"
#include "types.h"
#include "globals.h"
#include "alloc.h"
#include "hash.h"

extern DRAWSTATUS drawstatus;
extern SBOX defworld;
extern SBOX world;
extern SBOX windowport;
extern SBOX viewport;
extern PLANE *planes;

extern void draw_string ();
extern void draw_line ();
extern void draw_rectangle ();
extern void draw_circle ();
extern void draw_point ();
extern void plot_rectangle ();
extern void plot_circle ();
extern void plot_line ();
extern void plot_string ();
extern void redraw_all_planes ();
extern void init_pixmap (Pixmap, int, unsigned int, unsigned int);

int parse_quit (char *);
int parse_interactive (char *);
int parse_draw_line (char *);
int parse_draw_rectangle (char *);
int parse_fill_rectangle (char *);
int parse_draw_circle (char *);
int parse_fill_circle (char *);
int parse_draw_point (char *);
int parse_set_color (char *);
int parse_set_plane (char *);
int parse_set_font (char *);
int parse_draw_string (char *);
int parse_set_view_window (char *);
int parse_clear_plane (char *);
int parse_clear_planes (char *);
int parse_set_world (char *);
int parse_draw_planes (char *);
int parse_use_pixmap (char *);

static CMDLIST cmd_list[] = {
  { "clear-plane", parse_clear_plane },
  { "clear-planes", parse_clear_planes },
  { "draw-line", parse_draw_line },
  { "draw-point", parse_draw_point },
  { "draw-rectangle", parse_draw_rectangle },
  { "draw-string", parse_draw_string },
  { "fill-rectangle", parse_fill_rectangle },
  { "draw-all-planes", parse_draw_planes },
  { "draw-circle", parse_draw_circle },
  { "fill-circle", parse_fill_circle },
  { "interactive", parse_interactive },
  { "quit", parse_quit },
  { "set-color", parse_set_color },
/*  { "set-line-width", parse_set_linewidth }, to be implemented */
  { "set-font", parse_set_font },
  { "set-pixmap", parse_use_pixmap },
  { "set-plane", parse_set_plane },  
  { "set-view-port", parse_set_view_window },
  { "set-world", parse_set_world },
  { NULL, NULL }
};

static HASHTAB hashtab;
static HASHTAB fonttab;

extern void Delete_Object ();
extern void Insert_Object ();
extern void Print_Plane_Commands ();
extern void redraw_planes ();
extern void scale_box ();

void init_parse ()
{
  int i;
  hashtab  = make_hashtab (1);
  fonttab = make_hashtab (0);
  for (i = 0; cmd_list[i].name != NULL; i++) {
#ifdef DEBUG
    printf ("add cmd: %s\n", cmd_list[i].name);
#endif
    lookup (&hashtab, cmd_list[i].name, strlen (cmd_list[i].name), &cmd_list[i].function, INSERT);
  }
  
  /* add default font */
  lookup (&fonttab, "default", 7, &appdata.drawfont, INSERT);
}
		

void parse_token_and_exec (cmd, msg)
     char *cmd;
     char *msg;
{
  int (*func)();

  if (lookup  (&hashtab, cmd, strlen (cmd), &func, LOOKUP) == NOT_PRESENT) {
    if (appdata.verbose == True)
      fprintf (stderr, "ERROR: Command '%s' not found\n", cmd);
  }
  else {
    (*func)(msg);
  }
}

static char *skip_white (char *buf)
{
  while (*buf == ' ' || *buf == '\t')
    buf++;
  return buf;
}

static char *read_token (char *buf, char *token)
{
  char *stop;
  *token = '\0';
  while (isgraph(*buf)) {
    *token = *buf;
    buf++;
    token++;
  }
  *token = '\0';
  return buf;
}

/* return -1 on failure, 0 otherwise */

int parse_quit (msg)
     char *msg;
{
#ifdef DEBUG
  fprintf(stderr, "calling parse_quit [msg left: %s]\n", msg);
#endif
  exit (0);
}



int parse_draw_line (msg)
     char *msg;     
{
  int x1, x2, y1, y2;
  DRAWOBJ *obj;
  SBOX *box;
  LINE *line;
  
#ifdef DEBUG
  fprintf (stderr, "calling parse_draw_line [msg left: %s]\n", msg);
#endif

  if (sscanf(msg,"%d %d %d %d",&x1,&y1,&x2,&y2) != 4)
    return;
  box = NEW_STRUCT (SBOX);
  box->x = MIN (x1, x2);
  box->y = MIN (y1, y2);
  box->width = ABS (x1 - x2);
  box->height = ABS (y1 - y2);

  line = NEW_STRUCT (LINE);
  line->x1 = x1;
  line->y1 = y1;
  line->x2 = x2;
  line->y2 = y2;

  obj = NEW_STRUCT (DRAWOBJ);
  obj->sbox = box;
  obj->object = line;
  obj->drawfunc = draw_line;
  obj->plotfunc = plot_line;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}



int parse_draw_rectangle (msg)
     char *msg;

{
  int x, y, width, height;
  DRAWOBJ *obj;
  RECT *rect;
  SBOX *box;

#ifdef DEBUG
  fprintf (stderr, "calling parse_draw_rectangle [msg left: %s]\n", msg);
#endif
  if (sscanf(msg,"%d %d %d %d",&x,&y,&width,&height) !=4)
    return;  

  box = NEW_STRUCT (SBOX);
  box->x = x;
  box->y = y;
  box->width = width;
  box->height = height;
  rect = NEW_STRUCT (RECT);
  rect->fill = 0;
  obj = NEW_STRUCT (DRAWOBJ);
  obj->sbox = box;
  obj->object = rect;
  obj->drawfunc = draw_rectangle;
  obj->plotfunc = plot_rectangle;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}

  



int parse_fill_rectangle (msg)
     char *msg;
{
  int x, y, width, height;
  DRAWOBJ *obj;
  RECT *rect;
  SBOX *box;

#ifdef DEBUG
  fprintf (stderr, "calling parse_fill_rectangle [msg left: %s]\n",msg);
#endif

  if (sscanf(msg, "%d %d %d %d",&x,&y,&width,&height) !=4)
    return;

  box = NEW_STRUCT (SBOX);
  box->x = x;
  box->y = y;
  box->width = width;
  box->height = height;
  rect = NEW_STRUCT (RECT);
  rect->fill = 1;
  obj = NEW_STRUCT (DRAWOBJ);
  obj->sbox = box;
  obj->object = rect;
  obj->drawfunc = draw_rectangle;
  obj->plotfunc = plot_rectangle;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}



int parse_draw_circle (msg)
     char *msg;
{
  int cx,cy,radius;
  DRAWOBJ *obj;
  CIRCLE *circle;
  SBOX *box;

#ifdef DEBUG
  fprintf (stderr, "calling parse_draw_circle [msg left: %s]\n",msg);
#endif

  if (sscanf(msg, "%d %d %d",&cx,&cy,&radius) !=3)
    return;
  box = NEW_STRUCT (SBOX);
  box->x = cx - radius;
  box->y = cy - radius;
  box->width = 2 * radius;
  box->height = 2 * radius;
  circle = NEW_STRUCT (CIRCLE);
  circle->center_x = cx;
  circle->center_y = cy;
  circle->radius = radius;
  circle->fill = 0;
  obj = NEW_STRUCT (DRAWOBJ);
  obj->sbox = box;
  obj->object = circle;
  obj->drawfunc = draw_circle;
  obj->plotfunc = plot_circle;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}


  
int parse_fill_circle (msg)
     char *msg;
{
  int cx,cy,radius;
  DRAWOBJ *obj;
  CIRCLE *circle;
  SBOX *box;

#ifdef DEBUG
  fprintf (stderr, "calling parse_fill_circle [msg left: %s]\n",msg);
#endif
  
  if (sscanf (msg, "%d %d %d", &cx, &cy, &radius) != 3)
    return;

  box = NEW_STRUCT (SBOX);
  box->x = cx - radius;
  box->y = cy - radius;
  box->width = 2 * radius;
  box->height = 2 * radius;
  circle = NEW_STRUCT (CIRCLE);
  circle->center_x = cx;
  circle->center_y = cy;
  circle->radius = radius;
  circle->fill = 1;
  obj = NEW_STRUCT (DRAWOBJ);
  obj->sbox = box;
  obj->object = circle;
  obj->drawfunc = draw_circle;
  obj->plotfunc = plot_circle;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}


int parse_draw_point (msg)
     char *msg;
{
  DRAWOBJ *obj;
  SBOX *box;
  int x,y;

#ifdef DEBUG
  fprintf (stderr, "calling parse_draw_point [msg left: %s]\n",msg);
#endif

  if (sscanf(msg, "%d %d",&x,&y) !=2)
    return;

  box = NEW_STRUCT (SBOX);
  box->x = x;
  box->y = y;
  box->width = 0;
  box->height = 0;

  obj = NEW_STRUCT (DRAWOBJ);
  obj->sbox = box;
  obj->object = NULL;
  obj->drawfunc = draw_point;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}



int parse_set_color (msg)
     char *msg;
{
  static char colorname[128];
  Display *dpy  = XtDisplay(canvas);
  int      scr  = DefaultScreen(dpy);
  Colormap cmap = DefaultColormap(dpy, scr);
  XColor color;
  XColor ignore;
  Pixel pixel;
  XGCValues values;
  unsigned long mask;
  unsigned short red, green, blue;

#ifdef DEBUG  
  fprintf (stderr, "calling parse_set_color [msg left: %s]\n",msg);
#endif
  if (sscanf(msg, "%s", colorname) !=1)
    return;

  if (XParseColor(dpy, cmap, colorname, &color)) {

    if (XAllocColor(dpy, cmap, &color)) {
      pixel = color.pixel;
      red = color.red;
      green = color.green;
      blue = color.blue;

      /* set entry in color table */
      coltable[pixel] = color;
    }
    else {
      printf("Warning: Couldn't allocate color %s\n", colorname);
      pixel = BlackPixel(dpy, scr);
      red = green = blue = 0;
    }
  }
  else {
    printf("Warning: Couldn't allocate this #-argument\n");
  }

  values.foreground = pixel;
  mask = GCForeground;

  XChangeGC (XtDisplay (canvas), drawstatus.gcdata->gc, mask, &values);
  drawstatus.gcdata->def_pixel = pixel;
  drawstatus.gcdata->gc_pixel = pixel;
  drawstatus.gcdata->red = red;
  drawstatus.gcdata->green = green;
  drawstatus.gcdata->blue = blue;
  return 0;
}



int parse_set_linewidth (msg)
     char *msg;
{
  Display *dpy  = XtDisplay(canvas);
  int      scr  = DefaultScreen(dpy);
  XGCValues values;
  unsigned long mask;
  int lw;

#ifdef DEBUG  
  fprintf (stderr, "calling parse_set_linewidth [msg left: %s]\n",msg);
#endif
  if (sscanf(msg, "%d", &lw) !=1)
    return;

  return 0;
}


int parse_set_plane (msg)
     char *msg;
{
  int plane;
#ifdef DEBUG
  fprintf (stderr, "calling parse_set_plane [msg left: %s]\n",msg);
#endif
  if (sscanf(msg, "%d",&plane) !=1)
    return;

  if (plane < 0) {
    fprintf (stderr, "Plane number too small (%d); plane range = [%d,%d]\n",
	     plane, 0, appdata.nr_planes-1);
    return;
  }
  if (plane >= appdata.nr_planes) {
    fprintf (stderr, "Plane number too big (%d); plane range = [%d,%d]\n",
	     plane, 0, appdata.nr_planes-1);
    return;
  }
  drawstatus.plane = plane;
  if (appdata.plane_gc == True)
    drawstatus.gcdata = planes[plane].gcdata;
}  


int parse_set_font (char *msg)
{
  int key, insert;
  XFontStruct *fontstruct;
  char fontname[BUFSIZ];
  
  if (sscanf (msg, "%s", fontname) != 1)
    return;

  insert = INSERT;
  key = lookup (&fonttab, fontname, strlen (fontname), &fontstruct, &insert);

  if (insert != ALREADY_PRESENT) {
    fontstruct = XLoadQueryFont (XtDisplay (canvas), fontname);
    if ((fontstruct == NULL) && (appdata.verbose == True)) {
      fprintf (stderr, "Backdraw: Can not load font (%s)\n", fontname);
      return;
    }
    KEYINFO (fonttab, key) = (char *)fontstruct;
  }
  /* here fontstruct contains new font */
  if (fontstruct->fid !=  drawstatus.gcdata->def_font->fid) {
    XSetFont (XtDisplay (canvas), drawstatus.gcdata->gc, fontstruct->fid);
    drawstatus.gcdata->def_font = fontstruct;
    drawstatus.gcdata->gc_font = fontstruct;
  }
}


int parse_draw_string (char *msg)
{
  int x, y;
  char *start, *stop;
  SBOX *box;
  DRAWOBJ *obj;
  STRING *new;
  /* font info */
  int ascent, descent, direction;
  XCharStruct *charinfo;
  
  if (sscanf (msg, "%d %d", &x, &y) != 2)
    return;
  if ((start = strchr (msg, '\"')) == NULL)
    return;
  if ((stop = strchr (++start, '\"')) == NULL)
    return;
  *stop = '\0';


  new = NEW_STRUCT (STRING);
  new->x = x;
  new->y = y;
  new->str = NEW_STRING (start);
  new->font = drawstatus.gcdata->def_font;
  
  charinfo = NEW_STRUCT (XCharStruct);
  XTextExtents (drawstatus.gcdata->def_font, new->str, strlen (new->str),
		&direction, &ascent, &descent, charinfo);

  new->info = charinfo;

  box = NEW_STRUCT (SBOX);
  box->x = x + charinfo->lbearing * scaling;
  box->y = y - charinfo->ascent * scaling;
  box->width = (charinfo->lbearing + charinfo->rbearing) * scaling;
  box->height = (charinfo->ascent + charinfo->descent) * scaling;

  obj = NEW_STRUCT (DRAWOBJ);
  obj->type = DRAWOBJ_STRING;
  obj->sbox = box;
  obj->object = new;
  obj->drawfunc = draw_string;
  obj->plotfunc = plot_string;
  obj->pixel = drawstatus.gcdata->def_pixel;
  if (appdata.interactive == True) {
    draw_object (obj, drawstatus.gcdata, XtWindow (canvas));
    redraw_planes (drawstatus.plane + 1, box);
  }
  add_obj_to_plane (drawstatus.plane, obj);
}


int parse_set_view_window (char *msg)
{
  int x, y, width, height;
  float div_width, div_height;
  
  if (sscanf (msg, "%d %d %d %d", &x, &y, &width, &height) != 4)
    return;

  div_width = (float)width / windowport.width;
  div_height = (float)height / windowport.height;
  scaling = MAX (div_width, div_height);
  viewport.x = x;
  viewport.y = y;
  viewport.width = width;
  viewport.height = height;

  XClearArea (XtDisplay (canvas), XtWindow (canvas), 0, 0, 0, 0, TRUE);
}


int parse_clear_plane (char *msg)
{
  int plane;
  
  if (sscanf (msg, "%d", &plane) != 1)
    return;

  if ((plane >= 0) && (plane < appdata.nr_planes)) {
    clear_plane (plane);
    XClearArea (XtDisplay (canvas), XtWindow (canvas), 0, 0, 0, 0, TRUE);
  }
}


int parse_clear_planes (char *msg)
{
  int i;
  for (i = 0; i < appdata.nr_planes; i++)
    clear_plane (i);
  XClearArea (XtDisplay (canvas), XtWindow (canvas), 0, 0, 0, 0, TRUE);
}


int parse_set_world (char *msg)
{
  int x, y, width, height;
  
  if (sscanf(msg,"%d %d %d %d", &x, &y,
	     &width, &height) != 4)
    return;

  if ((width < 0) || (height < 0))
    return;
  
  defworld.x = x;
  defworld.y = y;
  defworld.width = width;
  defworld.height = height;
}

int parse_draw_planes (char *msg)
{
  redraw_all_planes ();
}


int parse_interactive (char *msg)
{
  int flag;

  if (appdata.update > 0)
    return;
  
  if (sscanf(msg,"%d", &flag) != 1)
    return;

  if (flag)
    appdata.interactive = True;
  else
    appdata.interactive = False;
}

int parse_use_pixmap (char *msg)
{
  int plane;
  
  if (sscanf(msg,"%d", &plane) != 1)
    return;

  if ((plane >= 0) && (plane < appdata.nr_planes)) {
    use_pixmap = plane;
    if (!pixmap) {
      Widget root;
      unsigned int x, y, width, height, border_width, depth;
      if (XGetGeometry (XtDisplay (canvas), XtWindow (canvas),
			 &root, &x, &y, &width, &height,
			 &border_width, &depth)) {
 	pixmap = XCreatePixmap (XtDisplay (canvas), XtWindow (canvas),
				width, height, depth);
	init_pixmap (pixmap, use_pixmap, width, height);
	pix_width = width;
	pix_height = height;
	pix_depth = depth;
      }
      else
	fprintf (stderr, "XGetGeometry failed\n");
    }
  }
  else {
    use_pixmap = -1;
    if (pixmap) {
      XFreePixmap (XtDisplay (canvas), pixmap);
      pixmap = NULL;
    }
  }
  
}
