#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <Xm/Xm.h>

#include "defines.h"
#include "types.h"
#include "plot.h"
#include "globals.h"
#include "alloc.h"

extern DRAWSTATUS drawstatus;
extern PLANE *planes;
extern PLOT_OPTIONS plot_options;
extern GC fillgc;

void init_planes (int no)
{
  int i;
  
  if (planes) {
    fprintf (stderr, "Planes are already initialised!\n");
    return;
  }

  if (no < 1) {
    fprintf (stderr, "No of planes should be a non-zero positive number\n");
    fprintf (stderr, "  set to: 1\n");
    no = 1;
  }

  if (no > MAX_PLANES) {
    fprintf (stderr, "No of planes should be a non-zero positive number\n");
    fprintf (stderr, "  set to: %d\n", MAX_PLANES);
    no = MAX_PLANES;
  }
  
  planes = NEW_ARRAY (no, PLANE);
  if (appdata.plane_gc == True) {
    for (i = 0; i < no; i++)
      planes[i].gcdata = NEW_STRUCT (GCDATA);
  }
}


void add_obj_to_plane (int no, DRAWOBJ *obj)
{
  obj->next = planes[no].startp;
  planes[no].startp = obj;
}


void redraw_all_planes ()
{
  int i;
  DRAWOBJ *obj;
  GCDATA *gcdata;

  if (appdata.plane_gc == False)
    gcdata = drawstatus.gcdata;
  for (i = 0; i < appdata.nr_planes; i++) {
    if (appdata.plane_gc == True)
      gcdata = planes[i].gcdata;
    for (obj = planes[i].startp; obj != NULL; obj = obj->next) {
      draw_object (obj, gcdata, XtWindow (canvas));
    }
    if (appdata.plane_gc == True)
      restore_gc (gcdata);
  }
  if (appdata.plane_gc == False)
    restore_gc (gcdata);
}


void redraw_planes (int from, SBOX *area)
{
  int i;
  DRAWOBJ *obj;
  GCDATA *gcdata;
  SBOX larea; /* local copy of area; this structure is updated */
  larea = *area;

  if (appdata.plane_gc == False)
    gcdata = drawstatus.gcdata;
  if ((use_pixmap >= 0) && (from <= use_pixmap)) {
    XCopyArea (XtDisplay (canvas), pixmap, XtWindow (canvas), gcdata->gc,
	       0, 0, pix_width, pix_height, 0, 0);
    from = use_pixmap + 1;
  }
  for (i = from; i < appdata.nr_planes; i++) {
    if (appdata.plane_gc == True)
      gcdata = planes[i].gcdata;
    for (obj = planes[i].startp; obj != NULL; obj = obj->next) {
      if (sbox_overlap (&larea, obj->sbox)) {
	update_sbox (&larea, obj->sbox);
#ifdef DEBUG
	fprintf (stderr, "area: %d %d %d %d\n", area->x, area->y,
		 area->x + area->width, area->y + area->height);
#endif
	draw_object (obj, gcdata, XtWindow (canvas));
      }
    }
    if (appdata.plane_gc == True)
      restore_gc (gcdata);
  }
  if (appdata.plane_gc == False)
    restore_gc (gcdata);
}


void resize_string_boxes (float factor)
{
  int i;
  DRAWOBJ *obj;
  for (i = 0; i < appdata.nr_planes; i++) {
    for (obj = planes[i].startp; obj != NULL; obj = obj->next) {
      if (obj->type == DRAWOBJ_STRING) {
	SBOX *sbox = obj->sbox;
	STRING *dstr = obj->object;
	XCharStruct *charinfo = dstr->info;
	sbox->x = dstr->x + charinfo->lbearing * factor;
	sbox->y = dstr->y - (charinfo->ascent * factor + 0.5);
	sbox->width = (charinfo->lbearing + charinfo->rbearing) * factor + 1;
	sbox->height = (charinfo->ascent + charinfo->descent) * factor + 1;
      }
    }
  }
}


void clear_plane (int plane)
{
  DRAWOBJ *ptr, *del;
  
  ptr = planes[plane].startp;
  planes[plane].startp = NULL;
  while (ptr) {
    del = ptr;
    ptr = ptr->next;
    FREE (del->sbox);
    FREE (del->object);
    FREE (del);
  }
}


void plot_planes (FILE *fp, SBOX *area)
{
  int i;
  DRAWOBJ *obj;
  SBOX larea; /* local copy of area; this structure is updated */
  larea = *area;

  plotpixel = nilpixel;
  for (i = 0; i < appdata.nr_planes; i++) {
    for (obj = planes[i].startp; obj != NULL; obj = obj->next) {
      if (sbox_overlap (&larea, obj->sbox)) {
	update_sbox (&larea, obj->sbox);
#ifdef DEBUG
	fprintf (stderr, "area: %d %d %d %d\n", area->x, area->y,
		 area->x + area->width, area->y + area->height);
#endif
	plot_object (obj, fp);
      }
    }
  }
}


/* 
 * =========================================================================
 * PIXMAP
 * =========================================================================
 */

void init_pixmap (Pixmap pixmap,
		  int max_plane,
		  unsigned int width,
		  unsigned height)
{
  int i;
  DRAWOBJ *obj;
  GCDATA *gcdata;

  /* copy background */
  XFillRectangle (XtDisplay (canvas), pixmap, fillgc, 0, 0, width, height);

  if (appdata.plane_gc == False)
    gcdata = drawstatus.gcdata;
  for (i = 0; i <= max_plane; i++) {
    if (appdata.plane_gc == True)
      gcdata = planes[i].gcdata;
    for (obj = planes[i].startp; obj != NULL; obj = obj->next) {
      draw_object (obj, gcdata, pixmap);
    }
    if (appdata.plane_gc == True)
      restore_gc (gcdata);
  }
  if (appdata.plane_gc == False)
    restore_gc (gcdata);
}
